from .version import __version__

from .binding import sort as _sort

import numpy
from numpy.lib.recfunctions import append_fields

try:
    unicode = unicode
except NameError:
    # 'unicode' is undefined, must be Python 3
    str = str
    unicode = str
    bytes = bytes
    basestring = (str,bytes)
else:
    # 'unicode' exists, must be Python 2
    str = str
    unicode = unicode
    bytes = str
    basestring = basestring

def sort(source, orderby=None, out=None, comm=None):
    """
        Sort source array with orderby as the key.
        Store result to out.

        Parameters
        ----------
        source : array, 1d, distributed

        orderby : array, 1d, distributed or string.
            Only integer types are supported.
            must be on the same partition as that of source.
            If orderby is string, it refers to the field in source.

        out : array, 1d distributed
            the total length must be the same as source.
            the itemsize must be the same as source
            if None, the sort is in-place.

        Returns
        -------
            out

        Remarks
        -------
            source, orderby, out can be flatiter at the same time.

    """

    key = orderby
    if isinstance(key, basestring):
        return _sort(source, key, out, comm=comm)

    if key is None:
        D, I = 'DD'
        data1 = numpy.empty(len(source),
                dtype=[('D', guess_dtype(source))])
        data1['D'][...] = source
    else:
        D, I = 'DI'
        data1 = numpy.empty(len(source),
                dtype=[('D', guess_dtype(source)),
                       ('I', guess_dtype(key))])

        data1['D'][...] = source
        data1['I'][...] = key

    if out is None:
        out = source
        _sort(data1, orderby=I, comm=comm)
        out[...] = data1[D][...]
    else:
        data2 = numpy.empty(len(out), dtype=data1.dtype)
        _sort(data1, orderby=I, out=data2, comm=comm)
        out[...] = data2[D][...]

    return out

def globalrange(array, comm):
    """
        The start and end of local chunk in the global array
    """
    s = comm.allgather(len(array))
    start = sum(s[:comm.rank])
    end = start + s[comm.rank]
    return (start, end)

def globalindices(array, comm):
    """
        Construct a list of indices in the global array.
    """
    start, end = globalrange(array, comm)
    globalsize = comm.bcast(end, root=comm.size - 1)

    if globalsize > 1024 * 1024 * 1024:
        dtype = 'i8'
    else:
        dtype = 'i4'

    return numpy.arange(start, end, dtype=dtype)

def guess_dtype(array):
    if hasattr(array, "dtype"):
        return array.dtype, array.shape[1:]
    else:
        return array[0].dtype, array[0].shape

def permute(source, argindex, comm, out=None):
    """
        Permute a distributed array.

        Constructs source[argindex], distributed as the same partition
        as argindex.

        Parameters
        ----------
        source : array, 1d, distributed
        argindex : array, 1d, distributed
        out : array 1d distributed for output

        Returns
        -------
            source[argindex] distributed as the same partition as argindex, or out.

    """

    source_size = comm.allreduce(len(source))
    argindex_size = comm.allreduce(len(argindex))

    if source_size != argindex_size:
        raise ValueError("Global size of source and argindex is different")

    if out is None:
        out = numpy.empty(len(argindex), guess_dtype(source))

    originind = globalindices(argindex, comm)
    originind2 = numpy.empty(len(source), dtype=originind.dtype)

    sort(originind, orderby=argindex, out=originind2, comm=comm)
    sort(source, orderby=originind2, out=out, comm=comm)
    return out

def histogram(array, bins, comm, right=False):
    """
        Histogram of array.

        Parameters
        ----------
        bins : collective, array, 1d.
            Bin edges.

        array : distributed, array, 1d, [] accepted.

        Returns
        -------
        The histogram. The bin edges.

    """
    if len(array) == 0:
        originrank = []
    else:
        originrank = numpy.digitize(array, bins, right)
    recv = numpy.bincount(originrank, minlength=len(bins) + 1)
    return comm.allreduce(recv)

def take(source, argindex, comm, out=None):
    """ Take argindex from source.

        source[argindex] distributed as argindex. argindex does not have to
        be a permutation, nor unique.

    """
    nlocal = len(source)
    start, end = globalrange(source, comm)

    # find number of selections made from my rank
    bins = comm.allgather(end)
    h = histogram(argindex, bins, comm)
    # nactive has number of selected objects from this rank.
    nactive = h[comm.rank]

    if out is None:
        out = numpy.empty(len(argindex), guess_dtype(source))

    originind = globalindices(argindex, comm)

    myargindex = numpy.empty(nactive, dtype=argindex.dtype)
    myoriginind = numpy.empty(nactive, dtype=originind.dtype)

    sort(originind, orderby=argindex, out=myoriginind, comm=comm)
    sort(argindex, orderby=argindex, out=myargindex, comm=comm)

    myresult = source[myargindex - start]

    sort(myresult, orderby=myoriginind, out=out, comm=comm)
    return out

from numpy.testing import Tester
test = Tester().test
bench = Tester().bench



import mpsort
from mpi4py import MPI
import numpy
from numpy.testing import assert_array_equal

def MPIWorld(NTask, required=[]):
    if not isinstance(NTask, (tuple, list)):
        NTask = (NTask,)

    if not isinstance(required, (tuple, list)):
        required = (required,)

    maxsize = max(required)
    if MPI.COMM_WORLD.size < maxsize:
        raise ValueError("Test Failed because the world is too small. Increase to mpirun -n %d" % maxsize)

    sizes = sorted(set(list(required) + list(NTask)))
    def dec(func):
        def wrapped(*args, **kwargs):
            for size in sizes:
                if MPI.COMM_WORLD.size < size: continue

                color = 0 if MPI.COMM_WORLD.rank < size else 1
                MPI.COMM_WORLD.barrier()
                comm = MPI.COMM_WORLD.Split(color)

                kwargs['comm'] = comm
                failed = 0
                msg = ""
                if color == 0:
                    assert comm.size == size
                    try:
                        func(*args, **kwargs)
                    except:
                        failed = 1
                        import traceback
                        msg = traceback.format_exc()
                gfailed = MPI.COMM_WORLD.allreduce(failed)
                if gfailed > 0:
                    msg = MPI.COMM_WORLD.allgather(msg)
                if failed: raise
                if gfailed > 0:
                    raise ValueError("Some ranks failed with %s" % "\n".join(msg))

        wrapped.__name__ = func.__name__
        return wrapped
    return dec

def split(array, comm, localsize=None):
    array = comm.bcast(array)
    if localsize is None:
        sp = numpy.array_split(array, comm.size)
        return comm.scatter(sp)
    else:
        g = comm.allgather(localsize)
        return comm.scatter(numpy.array_split(array, numpy.cumsum(g)[:-1]))

def heal(array, comm):
    a = comm.allgather(array)
    a = numpy.concatenate(a, axis=0)
    return a

def adjustsize(size, comm):
    ressize = size + 1 - 2 * ((comm.rank) % 2)
    if comm.size % 2 == 1:
        if comm.rank == 0:
            ressize = size
    return ressize

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_sort(comm):
    s = numpy.int32(numpy.random.random(size=1000) * 1000)

    local = split(s, comm)
    s = heal(local, comm)

    g = comm.allgather(local.size)
    mpsort.sort(local, orderby=None, out=None, comm=comm)

    r = heal(local, comm)
    s.sort()
    assert_array_equal(s, r)

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_sort_inplace(comm):
    s = numpy.int32(numpy.random.random(size=1000) * 1000)

    local = split(s, comm)
    s = heal(local, comm)

    g = comm.allgather(local.size)
    mpsort.sort(local, local, out=None, comm=comm)

    r = heal(local, comm)
    s.sort()
    assert_array_equal(s, r)

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_sort_outplace(comm):
    s = numpy.int32(numpy.random.random(size=1000) * 1000)

    local = split(s, comm)
    s = heal(local, comm)

    res = numpy.zeros(adjustsize(local.size, comm), dtype=local.dtype)

    mpsort.sort(local, local, out=res, comm=comm)

    s.sort()

    r = heal(res, comm)
    assert_array_equal(s, r)

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_sort_flatiter(comm):
    s = numpy.int32(numpy.random.random(size=1000) * 1000)

    local = split(s, comm)
    s = heal(local, comm)

    res = numpy.zeros(adjustsize(local.size, comm), dtype=local.dtype)

    mpsort.sort(local.flat, local.flat, out=res.flat, comm=comm)

    s.sort()

    r = heal(res, comm)
    assert_array_equal(s, r)

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_sort_struct(comm):
    s = numpy.empty(1000, dtype=[
        ('value', 'i8'),
        ('key', 'i8')])

    s['value'] = numpy.int32(numpy.random.random(size=1000) * 1000)
    s['key'] = s['value']

    local = split(s, comm)
    s = heal(local, comm)

    res = numpy.zeros_like(local)

    mpsort.sort(local, 'key', out=res, comm=comm)

    r = heal(res, comm)

    s.sort(order='key')
    assert_array_equal(s['value'], r['value'])

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_sort_struct_vector(comm):
    s = numpy.empty(10, dtype=[
        ('value', 'i8'),
        ('key', 'i8'),
        ('vkey', ('i8', 2))])

    s['value'] = numpy.int32(numpy.random.random(size=len(s)) * 1000)

    # use a scalar key to trick numpy
    # numpy sorts as byte streams for vector fields.
    s['key'][:][...] = s['value']
    s['vkey'][:, 0][...] = s['value']
    s['vkey'][:, 1][...] = s['value']

    local = split(s, comm)
    s = heal(local, comm)

    res = numpy.zeros_like(local)

    mpsort.sort(local, 'vkey', out=res, comm=comm)

    r = heal(res, comm)
    s.sort(order='key')
    assert_array_equal(s['value'], r['value'])

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_sort_vector(comm):
    s = numpy.empty(10, dtype=[('value', 'i8')])

    s['value'] = numpy.int32(numpy.random.random(size=len(s)) * 1000)

    local = split(s, comm)
    s = heal(local, comm)

    k = numpy.empty(len(local), ('i8', 2))
    k[:, 0][...] = local['value']
    k[:, 1][...] = local['value']

    res = numpy.zeros_like(local)

    mpsort.sort(local, k, out=res, comm=comm)

    s.sort(order='value')

    r = heal(res, comm)

    assert_array_equal(s['value'], r['value'])


@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_permute(comm):
    s = numpy.arange(10)
    local = split(s, comm)
    i = numpy.arange(9, -1, -1)
    ind = split(i, comm, adjustsize(local.size, comm))

    res = mpsort.permute(local, ind, comm)
    r = heal(res, comm)
    s = s[i]
    assert res.size == ind.size
    assert_array_equal(r, s)

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_permute_out(comm):
    s = numpy.arange(10)
    local = split(s, comm)
    i = numpy.arange(9, -1, -1)
    ind = split(i, comm, adjustsize(local.size, comm))

    res = numpy.empty(local.size, local.dtype)
    mpsort.permute(local, ind, comm, out=res)
    r = heal(res, comm)
    s = s[i]
    assert_array_equal(r, s)

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_take(comm):
    s = numpy.arange(10)
    local = split(s, comm)
    i = numpy.arange(9, -1, -1)
    ind = split(i, comm, adjustsize(local.size, comm))

    res = mpsort.take(local, ind, comm)
    r = heal(res, comm)
    s = s[i]
    assert res.size == ind.size
    assert_array_equal(r, s)

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_take_out(comm):
    s = numpy.arange(10)
    local = split(s, comm)
    i = numpy.arange(9, -1, -1)
    ind = split(i, comm, adjustsize(local.size, comm))

    res = numpy.empty(local.size, local.dtype)
    mpsort.take(local, ind, comm, out=res)
    r = heal(res, comm)
    s = s[i]
    assert_array_equal(r, s)

def test_version():
    import mpsort
    assert hasattr(mpsort, "__version__")

@MPIWorld(NTask=(1, 2, 3, 4), required=1)
def test_histogram_empty(comm):
    mpsort.histogram([], [1], comm)
    # no error shall be raised




#cython: embedsignature=True
cimport numpy
cimport libmpi as MPI
from libc.stddef cimport ptrdiff_t
from libc.stdint cimport uint64_t, int64_t, uint32_t, int32_t
from libc.string cimport memcpy
from libc.stdlib cimport abort
import numpy
from mpi4py import MPI as pyMPI

cdef extern from "radixsort.c":
    pass
cdef extern from "mpsort-mpi.c":
    void mpsort_mpi(void * base, size_t nmemb, size_t size,
            void (*radix)(void * ptr, void * radix, void * arg), 
            size_t rsize, 
            void * arg, MPI.MPI_Comm comm)
    void mpsort_mpi_newarray(void * base, size_t nmemb, 
            void * outbase, size_t outnmemb,
            size_t size,
            void (*radix)(void * ptr, void * radix, void * arg), 
            size_t rsize, 
            void * arg, MPI.MPI_Comm comm)

cdef struct MyClosure:
    int elsize
    void (*myradix)(const void * ptr, void * radix, void * arg) nogil
    ptrdiff_t radix_offset
    int radix_nmemb

cdef closure_init(MyClosure * self, numpy.dtype dtype, radixkey):
    cdef numpy.dtype radixdtype

    self.elsize = dtype.itemsize

    if radixkey is not None:
        radixdtype, self.radix_offset = dtype.fields[radixkey]
    else:
        radixdtype, self.radix_offset = dtype, 0

    if len(radixdtype.shape) == 0:
        self.radix_nmemb = 1
    elif len(radixdtype.shape) == 1:
        self.radix_nmemb = radixdtype.shape[0]
    else:
        raise ValueError("data[%s] is not 1d nor 2d %s" % (radixkey))

    #print 'radix offset =', self.radix_offset
    #print 'radix nmemb =', self.radix_nmemb
    #print 'radix dtype.shape = ', radixdtype.shape
    if radixdtype.base == numpy.dtype('u8'):
        self.myradix = myradix_u8
    elif radixdtype.base == numpy.dtype('i8'):
        self.myradix = myradix_i8
    elif radixdtype.base == numpy.dtype('u4'):
        self.myradix = myradix_u4
    elif radixdtype.base == numpy.dtype('i4'):
        self.myradix = myradix_i4
    else:
        raise TypeError("data[%s] is not u8 or i8" % (radixkey))

cdef void myradix_u8(const void * ptr, void * radix, void * arg) nogil:
    cdef MyClosure *clo = <MyClosure*> arg
    cdef char * rptr = <char*>radix
    cdef char * cptr = <char*> ptr
    cdef uint64_t value
    for i in range(clo.radix_nmemb):
        value = (<uint64_t *> (cptr + clo.radix_offset))[i]
        memcpy(rptr, &value, 8)
        rptr += 8

cdef void myradix_i8(const void * ptr, void * radix, void * arg) nogil:
    cdef MyClosure *clo = <MyClosure*> arg
    cdef char * rptr = <char*>radix
    cdef char * cptr = <char*> ptr
    cdef uint64_t value
    for i in range(clo.radix_nmemb):
        value = (<int64_t *> (cptr + clo.radix_offset))[i]
        value += <uint64_t> 9223372036854775808uL
        memcpy(rptr, &value, 8)
        rptr += 8

cdef void myradix_u4(const void * ptr, void * radix, void * arg) nogil:
    cdef MyClosure *clo = <MyClosure*> arg
    cdef char * rptr = <char*>radix
    cdef char * cptr = <char*> ptr
    cdef uint64_t value
    for i in range(clo.radix_nmemb):
        value = (<uint32_t *> (cptr + clo.radix_offset))[i]
        memcpy(rptr, &value, 8)
        rptr += 8

cdef void myradix_i4(const void * ptr, void * radix, void * arg) nogil:
    cdef MyClosure *clo = <MyClosure*> arg
    cdef char * rptr = <char*>radix
    cdef char * cptr = <char*> ptr
    cdef uint64_t value
    for i in range(clo.radix_nmemb):
        value = (<int32_t *> (cptr + clo.radix_offset))[i]
        value += <uint64_t> 9223372036854775808uL
        memcpy(rptr, &value, 8)
        rptr += 8

def sort(numpy.ndarray data, orderby=None, numpy.ndarray out=None, comm=None):
    """
        Parallel sort of distributed data set `data' over MPI Communicator `comm',
        ordered by key given in 'orderby'.

        data[orderby] must be of integer types.
        data[orderby] can be 2d, in which case the latter elements in a row has
        more significance.

        data must be C_contiguous numpy arrays,

        if orderby is None, use data itself.
    """
    cdef MyClosure clo
    cdef MPI.MPI_Comm mpicomm

    # assert you can access the orderby columns.
    key = data[orderby]

    if not data.flags['C_CONTIGUOUS']:
        raise ValueError("data must be C_CONTIGUOUS")

    if out is None:
        out = data

    if not out.flags['C_CONTIGUOUS']:
        raise ValueError("out must be C_CONTIGUOUS")

    if comm is None:
        comm = pyMPI.COMM_WORLD
        mpicomm = MPI.MPI_COMM_WORLD
    else:
        if isinstance(comm, pyMPI.Comm):
            if hasattr(pyMPI, '_addressof'):
                mpicomm = (<MPI.MPI_Comm*> (<numpy.intp_t>
                        pyMPI._addressof(comm))) [0]
            else:
                raise ValueError("only comm=None is supported, "
                        + " update mpi4py to a version with MPI._addressof")
        else:
            raise ValueError("only MPI.Comm objects are supported")

    Ntot = comm.allreduce(len(data))
    Ntotout = comm.allreduce(len(out))

    if Ntot != Ntotout:
        raise ValueError("total size of array changed %d != %d" % (Ntot, Ntotout))

    if data.dtype.itemsize != out.dtype.itemsize:
        raise ValueError("item size mismatch")

    closure_init(&clo, data.dtype, orderby)

    mpsort_mpi_newarray(data.data, len(data),
            out.data, len(out),
            data.dtype.itemsize, clo.myradix,
            clo.radix_nmemb * 8, <void*>&clo, mpicomm)




from distutils.core import setup
from Cython.Build import cythonize
from distutils.extension import Extension
import numpy

import mpi4py
import os
if 'MPICC' in os.environ:
    compiler = os.environ['MPICC']
else:
    try:
        compiler = str(mpi4py.get_config()['mpicc'])
    except:
        pass
    compiler = "mpicc"

os.environ['CC'] = compiler

if 'LDSHARED' not in os.environ:
    os.environ['LDSHARED'] = compiler + ' -shared'

extensions = [
        Extension("mpsort.binding", ["mpsort/binding.pyx"],
            include_dirs = ["./", numpy.get_include()])]

def find_version(path):
    import re
    # path shall be a plain ascii text file.
    s = open(path, 'rt').read()
    version_match = re.search(r"^__version__ = ['\"]([^'\"]*)['\"]",
                              s, re.M)
    if version_match:
        return version_match.group(1)
    raise RuntimeError("Version not found")

setup(
    name="mpsort",
    version=find_version("mpsort/version.py"),
    author="Yu Feng",
    author_email="rainwoodman@gmail.com",
    url="http://github.com/rainwoodman/mpsort",
    description="python binding of MP-sort, a peta scale sorting routine",
    zip_safe = False,
    package_dir = {'mpsort': 'mpsort'},
    install_requires=['cython', 'numpy'],
    packages= ['mpsort', 'mpsort.tests'],
    requires=['numpy'],
    ext_modules = cythonize(extensions)
)
